@@ -20,6 +20,7 @@ class EnvLog(object):
         cmd = [str(python_executable), VERSION_QUERY_SCRIPT]
         result = subprocess.check_output(cmd, universal_newlines=True)
         answer = json.loads(result)
+        answer["executable"] = python_executable
         self.dict["python"] = answer
 
     def get_commandlog(self, name):
@@ -35,7 +36,6 @@ class EnvLog(object):
         :param py.path.local installpkg: Path ot the package.
         """
         self.dict["installpkg"] = {
-            "md5": installpkg.computehash("md5"),
             "sha256": installpkg.computehash("sha256"),
             "basename": installpkg.basename,
         }
