@@ -7,9 +7,15 @@ import sys
 from itertools import chain
 
 import py
-from pkg_resources import to_filename
 
 import tox
+from tox import reporter
+from tox.action import Action
+from tox.config.parallel import ENV_VAR_KEY as PARALLEL_ENV_VAR_KEY
+from tox.constants import INFO, PARALLEL_RESULT_JSON_PREFIX, PARALLEL_RESULT_JSON_SUFFIX
+from tox.package.local import resolve_package
+from tox.util.lock import get_unique_file
+from tox.util.path import ensure_empty_dir
 
 from .config import DepConfig
 
@@ -17,7 +23,7 @@ from .config import DepConfig
 class CreationConfig:
     def __init__(
         self,
-        base_resolved_python_md5,
+        base_resolved_python_sha256,
         base_resolved_python_path,
         tox_version,
         sitepackages,
@@ -25,7 +31,7 @@ class CreationConfig:
         deps,
         alwayscopy,
     ):
-        self.base_resolved_python_md5 = base_resolved_python_md5
+        self.base_resolved_python_sha256 = base_resolved_python_sha256
         self.base_resolved_python_path = base_resolved_python_path
         self.tox_version = tox_version
         self.sitepackages = sitepackages
@@ -35,7 +41,7 @@ class CreationConfig:
 
     def writeconfig(self, path):
         lines = [
-            "{} {}".format(self.base_resolved_python_md5, self.base_resolved_python_path),
+            "{} {}".format(self.base_resolved_python_sha256, self.base_resolved_python_path),
             "{} {:d} {:d} {:d}".format(
                 self.tox_version, self.sitepackages, self.usedevelop, self.alwayscopy
             ),
@@ -58,11 +64,11 @@ class CreationConfig:
             alwayscopy = bool(int(alwayscopy))
             deps = []
             for line in lines:
-                base_resolved_python_md5, depstring = line.split(None, 1)
-                deps.append((base_resolved_python_md5, depstring))
-            base_resolved_python_md5, base_resolved_python_path = base_resolved_python_info
+                base_resolved_python_sha256, depstring = line.split(None, 1)
+                deps.append((base_resolved_python_sha256, depstring))
+            base_resolved_python_sha256, base_resolved_python_path = base_resolved_python_info
             return CreationConfig(
-                base_resolved_python_md5,
+                base_resolved_python_sha256,
                 base_resolved_python_path,
                 tox_version,
                 sitepackages,
@@ -75,7 +81,7 @@ class CreationConfig:
 
     def matches_with_reason(self, other, deps_matches_subset=False):
         for attr in (
-            "base_resolved_python_md5",
+            "base_resolved_python_sha256",
             "base_resolved_python_path",
             "tox_version",
             "sitepackages",
@@ -103,9 +109,36 @@ class CreationConfig:
 
 
 class VirtualEnv(object):
-    def __init__(self, envconfig=None, session=None):
+    def __init__(self, envconfig=None, popen=None, env_log=None):
         self.envconfig = envconfig
-        self.session = session
+        self.popen = popen
+        self._actions = []
+        self.env_log = env_log
+        self._result_json_path = None
+
+    def new_action(self, msg, *args):
+        config = self.envconfig.config
+        command_log = self.env_log.get_commandlog(
+            "test" if msg in ("run-test", "run-test-pre", "run-test-post") else "setup"
+        )
+        return Action(
+            self.name,
+            msg,
+            args,
+            self.envconfig.envlogdir,
+            config.option.resultjson,
+            command_log,
+            self.popen,
+            self.envconfig.envpython,
+        )
+
+    def get_result_json_path(self):
+        if self._result_json_path is None:
+            if self.envconfig.config.option.resultjson:
+                self._result_json_path = get_unique_file(
+                    self.path, PARALLEL_RESULT_JSON_PREFIX, PARALLEL_RESULT_JSON_SUFFIX
+                )
+        return self._result_json_path
 
     @property
     def hook(self):
@@ -149,7 +182,9 @@ class VirtualEnv(object):
             path = self._normal_lookup(name)
 
         if path is None:
-            raise tox.exception.InvocationError("could not find executable {!r}".format(name))
+            raise tox.exception.InvocationError(
+                "could not find executable {}".format(pipes.quote(name))
+            )
 
         return str(path)  # will not be rewritten for reporting
 
@@ -169,7 +204,7 @@ class VirtualEnv(object):
 
     def _check_external_allowed_and_warn(self, path):
         if not self.is_allowed_external(path):
-            self.session.report.warning(
+            reporter.warning(
                 "test command found but not installed in testenv\n"
                 "  cmd: {}\n"
                 "  env: {}\n"
@@ -231,11 +266,11 @@ class VirtualEnv(object):
         alwayscopy = self.envconfig.alwayscopy
         deps = []
         for dep in self.get_resolved_dependencies():
-            dep_name_md5 = getdigest(dep.name)
-            deps.append((dep_name_md5, dep.name))
-        base_resolved_python_md5 = getdigest(base_resolved_python_path)
+            dep_name_sha256 = getdigest(dep.name)
+            deps.append((dep_name_sha256, dep.name))
+        base_resolved_python_sha256 = getdigest(base_resolved_python_path)
         return CreationConfig(
-            base_resolved_python_md5,
+            base_resolved_python_sha256,
             base_resolved_python_path,
             version,
             sitepackages,
@@ -248,7 +283,7 @@ class VirtualEnv(object):
         dependencies = []
         for dependency in self.envconfig.deps:
             if dependency.indexserver is None:
-                package = self.session._resolve_package(package_spec=dependency.name)
+                package = resolve_package(package_spec=dependency.name)
                 if package != dependency.name:
                     dependency = dependency.__class__(package)
             dependencies.append(dependency)
@@ -265,9 +300,7 @@ class VirtualEnv(object):
         live_config = self._getliveconfig()
         if previous_config is None or not previous_config.matches(live_config):
             content = live_config.writeconfig(self.path_config)
-            self.session.report.verbosity1(
-                "write config to {} as {!r}".format(self.path_config, content)
-            )
+            reporter.verbosity1("write config to {} as {!r}".format(self.path_config, content))
 
     def _needs_reinstall(self, setupdir, action):
         setup_py = setupdir.join("setup.py")
@@ -290,7 +323,7 @@ class VirtualEnv(object):
             sys_path = json.loads(out)
         except ValueError:
             sys_path = []
-        egg_info_fname = ".".join((to_filename(name), "egg-info"))
+        egg_info_fname = ".".join((name.replace("-", "_"), "egg-info"))
         for d in reversed(sys_path):
             egg_info = py.path.local(d).join(egg_info_fname)
             if egg_info.check():
@@ -324,9 +357,9 @@ class VirtualEnv(object):
                 return
             action.setactivity("{}-nodeps".format(name), dir)
             pip_flags = ["--no-deps"] + ([] if is_develop else ["-U"])
-        pip_flags.extend(["-v"] * min(3, action.report.verbosity - 2))
-        if action.venv.envconfig.extras:
-            dir += "[{}]".format(",".join(action.venv.envconfig.extras))
+        pip_flags.extend(["-v"] * min(3, reporter.verbosity() - 2))
+        if self.envconfig.extras:
+            dir += "[{}]".format(",".join(self.envconfig.extras))
         target = [dir]
         if is_develop:
             target.insert(0, "-e")
@@ -360,7 +393,8 @@ class VirtualEnv(object):
 
         cmd = list(chain.from_iterable(expand(val) for val in self.envconfig.install_command))
 
-        self.ensure_pip_os_environ_ok()
+        env = self._get_os_environ()
+        self.ensure_pip_os_environ_ok(env)
 
         old_stdout = sys.stdout
         sys.stdout = codecs.getwriter("utf8")(sys.stdout)
@@ -369,28 +403,33 @@ class VirtualEnv(object):
                 cmd,
                 cwd=self.envconfig.config.toxinidir,
                 action=action,
-                redirect=self.session.report.verbosity < 2,
+                redirect=reporter.verbosity() < reporter.Verbosity.DEBUG,
+                env=env,
             )
         finally:
             sys.stdout = old_stdout
 
-    def ensure_pip_os_environ_ok(self):
+    def ensure_pip_os_environ_ok(self, env):
         for key in ("PIP_RESPECT_VIRTUALENV", "PIP_REQUIRE_VIRTUALENV", "__PYVENV_LAUNCHER__"):
-            os.environ.pop(key, None)
-        if "PYTHONPATH" not in self.envconfig.passenv:
+            env.pop(key, None)
+        if all("PYTHONPATH" not in i for i in (self.envconfig.passenv, self.envconfig.setenv)):
             # If PYTHONPATH not explicitly asked for, remove it.
-            if "PYTHONPATH" in os.environ:
-                self.session.report.warning(
-                    "Discarding $PYTHONPATH from environment, to override "
-                    "specify PYTHONPATH in 'passenv' in your configuration."
-                )
-                os.environ.pop("PYTHONPATH")
+            if "PYTHONPATH" in env:
+                if sys.version_info < (3, 4) or bool(env["PYTHONPATH"]):
+                    # https://docs.python.org/3/whatsnew/3.4.html#changes-in-python-command-behavior
+                    # In a posix shell, setting the PATH environment variable to an empty value is
+                    # equivalent to not setting it at all.
+                    reporter.warning(
+                        "Discarding $PYTHONPATH from environment, to override "
+                        "specify PYTHONPATH in 'passenv' in your configuration."
+                    )
+                env.pop("PYTHONPATH")
 
         # installing packages at user level may mean we're not installing inside the venv
-        os.environ["PIP_USER"] = "0"
+        env["PIP_USER"] = "0"
 
         # installing without dependencies may lead to broken packages
-        os.environ["PIP_NO_DEPS"] = "0"
+        env["PIP_NO_DEPS"] = "0"
 
     def _install(self, deps, extraopts=None, action=None):
         if not deps:
@@ -438,7 +477,7 @@ class VirtualEnv(object):
     def test(
         self,
         redirect=False,
-        name="runtests",
+        name="run-test",
         commands=None,
         ignore_outcome=None,
         ignore_errors=None,
@@ -450,7 +489,7 @@ class VirtualEnv(object):
             ignore_outcome = self.envconfig.ignore_outcome
         if ignore_errors is None:
             ignore_errors = self.envconfig.ignore_errors
-        with self.session.newaction(self, name) as action:
+        with self.new_action(name) as action:
             cwd = self.envconfig.changedir
             if display_hash_seed:
                 env = self._get_os_environ(is_test_command=True)
@@ -486,17 +525,16 @@ class VirtualEnv(object):
                 except tox.exception.InvocationError as err:
                     if ignore_outcome:
                         msg = "command failed but result from testenv is ignored\ncmd:"
-                        self.session.report.warning("{} {}".format(msg, err))
+                        reporter.warning("{} {}".format(msg, err))
                         self.status = "ignored failed command"
                         continue  # keep processing commands
 
-                    self.session.report.error(str(err))
+                    reporter.error(str(err))
                     self.status = "commands failed"
                     if not ignore_errors:
                         break  # Don't process remaining commands
                 except KeyboardInterrupt:
                     self.status = "keyboardinterrupt"
-                    self.session.report.error(self.status)
                     raise
 
     def _pcall(
@@ -509,13 +547,16 @@ class VirtualEnv(object):
         redirect=True,
         ignore_ret=False,
         returnout=False,
+        env=None,
     ):
+        if env is None:
+            env = self._get_os_environ(is_test_command=is_test_command)
+
         # construct environment variables
-        os.environ.pop("VIRTUALENV_PYTHON", None)
-        env = self._get_os_environ(is_test_command=is_test_command)
+        env.pop("VIRTUALENV_PYTHON", None)
         bin_dir = str(self.envconfig.envbindir)
         env["PATH"] = os.pathsep.join([bin_dir, os.environ["PATH"]])
-        self.session.report.verbosity2("setting PATH={}".format(env["PATH"]))
+        reporter.verbosity2("setting PATH={}".format(env["PATH"]))
 
         # get command
         args[0] = self.getcommandpath(args[0], venv, cwd)
@@ -524,15 +565,71 @@ class VirtualEnv(object):
 
         cwd.ensure(dir=1)  # ensure the cwd exists
         return action.popen(
-            args, cwd=cwd, env=env, redirect=redirect, ignore_ret=ignore_ret, returnout=returnout
+            args,
+            cwd=cwd,
+            env=env,
+            redirect=redirect,
+            ignore_ret=ignore_ret,
+            returnout=returnout,
+            report_fail=not is_test_command,
         )
 
+    def setupenv(self):
+        if self.envconfig._missing_subs:
+            self.status = (
+                "unresolvable substitution(s): {}. "
+                "Environment variables are missing or defined recursively.".format(
+                    ",".join(["'{}'".format(m) for m in self.envconfig._missing_subs])
+                )
+            )
+            return
+        if not self.matching_platform():
+            self.status = "platform mismatch"
+            return  # we simply omit non-matching platforms
+        with self.new_action("getenv", self.envconfig.envdir) as action:
+            self.status = 0
+            default_ret_code = 1
+            envlog = self.env_log
+            try:
+                status = self.update(action=action)
+            except IOError as e:
+                if e.args[0] != 2:
+                    raise
+                status = (
+                    "Error creating virtualenv. Note that spaces in paths are "
+                    "not supported by virtualenv. Error details: {!r}".format(e)
+                )
+            except tox.exception.InvocationError as e:
+                status = e
+            except tox.exception.InterpreterNotFound as e:
+                status = e
+                if self.envconfig.config.option.skip_missing_interpreters == "true":
+                    default_ret_code = 0
+            if status:
+                str_status = str(status)
+                command_log = envlog.get_commandlog("setup")
+                command_log.add_command(["setup virtualenv"], str_status, default_ret_code)
+                self.status = status
+                if default_ret_code == 0:
+                    reporter.skip(str_status)
+                else:
+                    reporter.error(str_status)
+                return False
+            command_path = self.getcommandpath("python")
+            envlog.set_python_info(command_path)
+            return True
+
+    def finishvenv(self):
+        with self.new_action("finishvenv"):
+            self.finish()
+            return True
+
 
 def getdigest(path):
     path = py.path.local(path)
     if not path.check(file=1):
         return "0" * 32
-    return path.computehash()
+    return path.computehash("sha256")
 
 
 def prepend_shebang_interpreter(args):
@@ -559,7 +656,7 @@ def prepend_shebang_interpreter(args):
     return args
 
 
-NO_DOWNLOAD = False
+_SKIP_VENV_CREATION = os.environ.get("_TOX_SKIP_ENV_CREATION_TEST", False) == "1"
 
 
 @tox.hookimpl
@@ -570,18 +667,72 @@ def tox_testenv_create(venv, action):
         args.append("--system-site-packages")
     if venv.envconfig.alwayscopy:
         args.append("--always-copy")
-    if NO_DOWNLOAD:
+    if not venv.envconfig.download:
         args.append("--no-download")
     # add interpreter explicitly, to prevent using default (virtualenv.ini)
     args.extend(["--python", str(config_interpreter)])
-    venv.session.make_emptydir(venv.path)
-    basepath = venv.path.dirpath()
-    basepath.ensure(dir=1)
+
+    cleanup_for_venv(venv)
+
+    base_path = venv.path.dirpath()
+    base_path.ensure(dir=1)
     args.append(venv.path.basename)
-    venv._pcall(args, venv=False, action=action, cwd=basepath)
+    if not _SKIP_VENV_CREATION:
+        try:
+            venv._pcall(
+                args,
+                venv=False,
+                action=action,
+                cwd=base_path,
+                redirect=reporter.verbosity() < reporter.Verbosity.DEBUG,
+            )
+        except KeyboardInterrupt:
+            venv.status = "keyboardinterrupt"
+            raise
     return True  # Return non-None to indicate plugin has completed
 
 
+def cleanup_for_venv(venv):
+    within_parallel = PARALLEL_ENV_VAR_KEY in os.environ
+    # if the directory exists and it doesn't look like a virtualenv, produce
+    # an error
+    if venv.path.exists():
+        dir_items = set(os.listdir(str(venv.path))) - {".lock", "log"}
+        dir_items = {p for p in dir_items if not p.startswith(".tox-") or p == ".tox-config1"}
+    else:
+        dir_items = set()
+
+    if not (
+        # doesn't exist => OK
+        not venv.path.exists()
+        # does exist, but it's empty => OK
+        or not dir_items
+        # tox has marked this as an environment it has created in the past
+        or ".tox-config1" in dir_items
+        # it exists and we're on windows with Lib and Scripts => OK
+        or (INFO.IS_WIN and dir_items > {"Scripts", "Lib"})
+        # non-windows, with lib and bin => OK
+        or dir_items > {"bin", "lib"}
+        # pypy has a different lib folder => OK
+        or dir_items > {"bin", "lib_pypy"}
+    ):
+        venv.status = "error"
+        reporter.error(
+            "cowardly refusing to delete `envdir` (it does not look like a virtualenv): "
+            "{}".format(venv.path)
+        )
+        raise SystemExit(2)
+
+    if within_parallel:
+        if venv.path.exists():
+            # do not delete the log folder as that's used by parent
+            for content in venv.path.listdir():
+                if not content.basename == "log":
+                    content.remove(rec=1, ignore_errors=True)
+    else:
+        ensure_empty_dir(venv.path)
+
+
 @tox.hookimpl
 def tox_testenv_install_deps(venv, action):
     deps = venv.get_resolved_dependencies()
@@ -601,7 +752,7 @@ def tox_runtest(venv, redirect):
 @tox.hookimpl
 def tox_runtest_pre(venv):
     venv.status = 0
-    venv.session.make_emptydir(venv.envconfig.envtmpdir)
+    ensure_empty_dir(venv.envconfig.envtmpdir)
     venv.envconfig.envtmpdir.ensure(dir=1)
     venv.test(
         name="run-test-pre",
@@ -628,7 +779,7 @@ def tox_runtest_post(venv):
 def tox_runenvreport(venv, action):
     # write out version dependency information
     args = venv.envconfig.list_dependencies_command
-    output = venv._pcall(args, cwd=venv.envconfig.config.toxinidir, action=action)
+    output = venv._pcall(args, cwd=venv.envconfig.config.toxinidir, action=action, returnout=True)
     # the output contains a mime-header, skip it
     output = output.split("\n\n")[-1]
     packages = output.strip().split("\n")
